/* This file is part of mailfrom filter.
   Copyright (C) 2005, 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <resolv.h>

#include "mailfrom.h"

struct mx_buffer {
	unsigned pref;
	char *name;
};

static int
comp_pref(const void *a, const void *b)
{
	const struct mx_buffer *ma = a, *mb = b;
	if (ma->pref > mb->pref)
		return 1;
	else if (ma->pref < mb->pref)
		return -1;
	return 0;
}

/* Obtain MX records for domain name HOST. Return them in mxbuf, sorted by
   preference in ascending order.
   Notice that we *have* to use old BIND-4-style method, since glibc folks
   in their incredible wiseness have not exported new interface functions
   from libresolve.so. The only way to access them would be to link with
   -lresolve statically.
   
   Let's hope that someday this stupidity will change. */
   
static mf_status
_getmx(char *host, char *answer, size_t answer_size, mxbuf_t mxbuf)
{
	int i, n, nmx;
	struct mx_buffer mx_buffer[MAXMXCOUNT];
	HEADER *hp;
	unsigned char *eom, *cp;
	unsigned short qdcount, ancount;
	struct __res_state stat;

	debug1(80,"Getting MX records for %s", host);
	memset((void *)&stat, 0, sizeof(struct __res_state));
	res_ninit(&stat);
	n = res_nquery (&stat, host, C_IN, T_MX, answer, answer_size);
	res_nclose(&stat);
	if (n < 0) {
		debug2(10, "res_nquery failed (errno=%s, h_errno=%d)",
			   mu_strerror(errno), h_errno);
		switch (h_errno) {
		case NO_DATA:
		case NO_RECOVERY:
		case HOST_NOT_FOUND:
			return mf_not_found;

		case TRY_AGAIN:
		case -1:
			return mf_temp_failure;

		default:
			mu_error("res_nquery(%s) failed with unexpected h_errno %d",
				 host, h_errno);
			return mf_failure;
		}
	}

	if (n > answer_size)
		n = answer_size;

	hp = (HEADER*)answer;
	cp = (unsigned char *) answer + HFIXEDSZ;
	eom = (unsigned char *) answer + n;
	
	/* Skip query part */
	for (qdcount = ntohs((unsigned short)hp->qdcount);
	     qdcount--;
	     cp += n + QFIXEDSZ)
	{
		if ((n = dn_skipname(cp, eom)) < 0)
			return mf_failure;
	}

	ancount = ntohs((unsigned short)hp->ancount);

	/* Collect MX records */
	for (i = nmx = 0; i < ancount; i++) {
		unsigned short pref, type;
		char tname[NS_MAXDNAME];

		if ((n = dn_expand((unsigned char *)answer,
				   eom, cp, tname, sizeof tname)) < 0)
			break;
		cp += n;
		GETSHORT(type, cp);
		cp += INT16SZ + INT32SZ;
		GETSHORT(n, cp);
		if (type != T_MX) {
			debug2(90,"unexpected answer type %d, size %d\n",
				   type, n);
			cp += n;
			continue;
		}
		GETSHORT(pref, cp);
		if ((n = dn_expand((u_char *)answer, eom, cp,
				   tname, sizeof tname)) < 0)
			break;
		cp += n;
		mx_buffer[nmx].pref = pref;
		mx_buffer[nmx].name = strdup(tname);
		debug2(20,"MX %u %s", mx_buffer[nmx].pref,
			   mx_buffer[nmx].name);
		if (++nmx >= MAXMXCOUNT)
			break;
	}

	/* Sort according to preference value */
	qsort(mx_buffer, nmx, sizeof mx_buffer[0], comp_pref);

	/* Prepare return value */
	memset(mxbuf, 0, sizeof(mxbuf_t));
	for (i = 0; i < nmx; i++) 
		mxbuf[i] = mx_buffer[i].name;
	return mf_success;
}

static int
is_ipaddr(char *str)
{
	for (; *str; str++)
		if (!(isdigit(*str) || *str == '.'))
			return 0;
	return 1;
}

MUTEX_DCL(mf_mutex)

/* Return MX records for the given HOST. If no records were found, recurse
   to its parent domains until any record is found or recursion depth reaches
   MAXDEPTH */
static mf_status
get_mx_records(char *host, mxbuf_t mxbuf, int maxdepth)
{
	char hbuf[NI_MAXHOST];
	mf_status status = mf_failure;
	
	MUTEX_LOCK(mf_mutex);
	if (is_ipaddr(host)) {
		struct sockaddr_in sa;
		memset(&sa, 0, sizeof sa);
		sa.sin_family = AF_INET;
		if (inet_aton(host, &sa.sin_addr)
		    && getnameinfo((struct sockaddr *)&sa, sizeof sa,
				   hbuf, sizeof(hbuf),
				   NULL, 0,
				   0) == 0
		    && !is_ipaddr(hbuf))
			host = hbuf;
		else
			host = NULL;
	}
	
	if (host) {		
		unsigned char *answer = malloc(MAXPACKET);
		if (!answer)
			status = mf_failure;
		else {
			char *p;
			int depth;
			
			for (p = host, depth = 0; p && depth < maxdepth;
			     p++, depth++) {
				status = _getmx(p, answer, MAXPACKET, mxbuf);
				if (status == mf_success 
				    || status == mf_temp_failure)
					break;
				p = strchr(p, '.');
				if (!p)
					break;
			}
			free(answer);
		}
	}
	MUTEX_UNLOCK(mf_mutex);
	return status;
}

mf_status
getmx(char *host, mxbuf_t mxbuf)
{
	return get_mx_records(host, mxbuf, 1);
}

void
freemx(mxbuf_t mxbuf)
{
	int i;
	for (i = 0; i < MAXMXCOUNT && mxbuf[i]; i++)
		free(mxbuf[i]);
}

#define MAXCNAMEDEPTH 24 /* FIXME */

#define LOOKUP_FAILURE  -1
#define LOOKUP_SUCCESS 0
#define LOOKUP_CNAME   1

struct loop_data {
	char *name;
	size_t name_size;
	char *domain;
	size_t domain_size;
	char *answer;
	size_t answer_size;
	char *hbuf;
	size_t hbsize;
	mf_status status;
	size_t loopcnt;
};

#define NSIZE MAX(MAXPACKET, MAXDNAME*2+2)
#define SET_STATUS(lp,s) if ((lp)->status != mf_success) (lp)->status = s

int
cnameloop(struct loop_data *lp)
{
	int i, n, rc;
	HEADER *hp;
	unsigned char *eom, *cp;
	unsigned short qdcount, ancount;
	char *p;
	size_t len;
	struct __res_state statb;

	memset((void *)&statb, 0, sizeof(struct __res_state));
	res_ninit(&statb);
	n = res_nquerydomain (&statb, lp->name, lp->domain, C_IN, T_ANY,
			     lp->answer, lp->answer_size);
	res_nclose(&statb);
	if (n < 0) {
		debug2(10, "res_nquerydomain failed (errno=%s, h_errno=%d)",
		       mu_strerror(errno), h_errno);
		switch (h_errno) {
		case NO_DATA:
		case NO_RECOVERY:
		case HOST_NOT_FOUND:
			SET_STATUS(lp, mf_not_found);
			return LOOKUP_FAILURE;

		case TRY_AGAIN:
		case -1:
			SET_STATUS(lp, mf_temp_failure);
			return LOOKUP_FAILURE;

		default:
			mu_error("res_nquerydomain(%s) failed with unexpected h_errno %d",
				 lp->name, h_errno);
			SET_STATUS(lp, mf_failure);
			return LOOKUP_FAILURE;
		}
	}

	if (n > lp->answer_size)
		n = lp->answer_size;

	hp = (HEADER*) lp->answer;
	cp = (unsigned char *) lp->answer + HFIXEDSZ;
	eom = (unsigned char *) lp->answer + n;
	
	/* Skip query part */
	for (qdcount = ntohs((unsigned short)hp->qdcount);
	     qdcount--;
	     cp += n + QFIXEDSZ) {
		if ((n = dn_skipname(cp, eom)) < 0) {
			SET_STATUS(lp, mf_failure);
			return LOOKUP_FAILURE;
		}
	}

	for (ancount = ntohs((unsigned short) hp->ancount), i = 0;
	     i < ancount && cp < eom;
	     cp += n, i++) {
		unsigned short type;
		int ttl = 0;
		char nbuf[NSIZE];
		
		n = dn_expand((unsigned char *) lp->answer, eom, cp,
			      nbuf, sizeof nbuf);
		if (n < 0)
			break;
		cp += n;
		GETSHORT(type, cp);
		cp += INT16SZ;		/* skip over class */
		GETLONG(ttl, cp);
		GETSHORT(n, cp);	/* rdlength */
		switch (type) {
		case T_PTR:
			if ((rc = dn_expand((unsigned char *)lp->answer,
					    eom, cp,
					    nbuf, sizeof(nbuf))) < 0) {
				SET_STATUS(lp, mf_failure);
				return LOOKUP_FAILURE;
			}
			
			strncpy(lp->hbuf, nbuf, lp->hbsize);
			lp->hbuf[lp->hbsize-1] = 0;
			SET_STATUS(lp, mf_success);
			return LOOKUP_SUCCESS;
			
		case T_CNAME:
			if (--lp->loopcnt == 0) {
				mu_error ("DNS failure: CNAME loop for %s",
					  lp->name);
				SET_STATUS(lp, mf_failure);
				return LOOKUP_FAILURE;
			}

			if ((rc = dn_expand((unsigned char *)lp->answer,
					    eom, cp,
					    nbuf, sizeof(nbuf))) < 0) {
				SET_STATUS(lp, mf_failure);
				return LOOKUP_FAILURE;
			}
			
			strncpy(lp->hbuf, nbuf, lp->hbsize);
			lp->hbuf[lp->hbsize-1] = 0;
			/* RFC 1034 section 3.6 specifies that CNAME
			   should point at the canonical name -- but
			   urges software to try again anyway.
			*/
			p = strchr (lp->hbuf, '.');
			if (!p)
				return LOOKUP_SUCCESS;
			len = p - lp->hbuf;
			if (len + 1 >= lp->name_size)
				return LOOKUP_FAILURE;
			memcpy(lp->name, lp->hbuf, len);
			lp->name[len] = 0;
			len = strlen (p + 1);
			if (len + 1 >= lp->domain_size)
				return LOOKUP_FAILURE;
			strcpy(lp->domain, p + 1);
			SET_STATUS(lp, mf_success);
			return LOOKUP_CNAME;
		}
	}	
	SET_STATUS(lp, mf_not_found);
	return LOOKUP_FAILURE;
}

mf_status
_resolve_ipstr(char *ipstr, char *answer, size_t answer_size,
	       char *hbuf, size_t hbsize)
{
	int i;
	char revipstr[3*4+3+1];
	char namebuf[NSIZE];
	char domainbuf[NSIZE];
	char *p, *q;
	struct loop_data ld;

	q = revipstr + strlen (ipstr);
	*q = 0;
	for (i = 0, p = ipstr; *p && i < 4; i++) {
		int len;

		for (len = 0; p[len] && p[len] != '.'; len++)
			;
		q -= len;
		memcpy(q, p, len);
		if (q > revipstr)
			*--q = '.';
		p += len;
		if (*p == '.')
			p++;
	}

	if (*p || i != 4)
		return mf_failure;
		
	ld.name = namebuf;
	ld.name_size = sizeof namebuf;
	strcpy(namebuf, revipstr);
	ld.domain = domainbuf;
	strcpy(domainbuf, "in-addr.arpa");
	ld.domain_size = sizeof domainbuf;
	ld.answer = answer;
	ld.answer_size = answer_size;
	ld.hbuf = hbuf;
	ld.hbsize = hbsize;
	ld.status = mf_failure;
	ld.loopcnt = MAXCNAMEDEPTH;

	while (cnameloop(&ld) == LOOKUP_CNAME)
		;
	return ld.status;
}

mf_status
resolve_ipstr(char *ipstr, char *hbuf, size_t hbsize)	
{
	mf_status status;
	static unsigned char *answer;

	if (!answer)
		answer = malloc(MAXPACKET);
	debug1(80,"Getting canonical name for %s", ipstr);
	MUTEX_LOCK(mf_mutex);
	status = _resolve_ipstr(ipstr, answer, MAXPACKET, hbuf, hbsize);
	MUTEX_UNLOCK(mf_mutex);
	if (status == mf_success)
		debug2(80, "%s resolved to %s", ipstr, hbuf);
	else
		debug1(80, "%s not resolved", ipstr);
	return status;
}
	

